from array import array
from bisect import bisect_right
import io
from os import PathLike
from os.path import isfile
from struct import Struct
import sys
from typing import BinaryIO, ClassVar, Literal, cast
import warnings

from pyzstd import (
    _DEPRECATED_PLACEHOLDER,
    ZstdCompressor,
    ZstdDecompressor,
    _DeprecatedPlaceholder,
    _LevelOrOption,
    _Option,
    _StrOrBytesPath,
    _ZstdDict,
)

if sys.version_info < (3, 12):
    from typing_extensions import Buffer
else:
    from collections.abc import Buffer

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

__all__ = ("SeekableFormatError", "SeekableZstdFile")

_MODE_CLOSED = 0
_MODE_READ = 1
_MODE_WRITE = 2


class SeekableFormatError(Exception):
    "An error related to Zstandard Seekable Format."

    def __init__(self, msg: str) -> None:
        super().__init__("Zstandard Seekable Format error: " + msg)


__doc__ = """\
Zstandard Seekable Format (Ver 0.1.0, Apr 2017)
Square brackets are used to indicate optional fields.
All numeric fields are little-endian unless specified otherwise.
A. Seek table is a skippable frame at the end of file:
     Magic_Number  Frame_Size  [Seek_Table_Entries]  Seek_Table_Footer
     4 bytes       4 bytes     8-12 bytes each       9 bytes
     Magic_Number must be 0x184D2A5E.
B. Seek_Table_Entries:
     Compressed_Size  Decompressed_Size  [Checksum]
     4 bytes          4 bytes            4 bytes
     Checksum is optional.
C. Seek_Table_Footer:
     Number_Of_Frames  Seek_Table_Descriptor  Seekable_Magic_Number
     4 bytes           1 byte                 4 bytes
     Seekable_Magic_Number must be 0x8F92EAB1.
D. Seek_Table_Descriptor:
     Bit_number  Field_name
     7           Checksum_Flag
     6-2         Reserved_Bits  (should ensure they are set to 0)
     1-0         Unused_Bits    (should not interpret these bits)"""
__format_version__ = "0.1.0"


class _SeekTable:
    _s_2uint32 = Struct("<II")
    _s_3uint32 = Struct("<III")
    _s_footer = Struct("<IBI")

    # read_mode is True for read mode, False for write/append modes.
    def __init__(self, *, read_mode: bool) -> None:
        self._read_mode = read_mode
        self._clear_seek_table()

    def _clear_seek_table(self) -> None:
        self._has_checksum = False
        # The seek table frame size, used for append mode.
        self._seek_frame_size = 0
        # The file size, used for seeking to EOF.
        self._file_size = 0

        self._frames_count = 0
        self._full_c_size = 0
        self._full_d_size = 0

        if self._read_mode:
            # Item: cumulated_size
            # Length: frames_count + 1
            # q is int64_t. On Linux/macOS/Windows, Py_off_t is signed, so
            # ZstdFile/SeekableZstdFile use int64_t as file position/size.
            self._cumulated_c_size = array("q", [0])
            self._cumulated_d_size = array("q", [0])
        else:
            # Item: (c_size1, d_size1,
            #        c_size2, d_size2,
            #        c_size3, d_size3,
            #        ...)
            # Length: frames_count * 2
            # I is uint32_t.
            self._frames = array("I")

    def append_entry(self, compressed_size: int, decompressed_size: int) -> None:
        if compressed_size == 0:
            if decompressed_size == 0:
                # (0, 0) frame is no sense
                return
            # Impossible frame
            raise ValueError

        self._frames_count += 1
        self._full_c_size += compressed_size
        self._full_d_size += decompressed_size

        if self._read_mode:
            self._cumulated_c_size.append(self._full_c_size)
            self._cumulated_d_size.append(self._full_d_size)
        else:
            self._frames.append(compressed_size)
            self._frames.append(decompressed_size)

    # seek_to_0 is True or False.
    # In read mode, seeking to 0 is necessary.
    def load_seek_table(self, fp: BinaryIO, seek_to_0: bool) -> None:  # noqa: FBT001
        # Get file size
        fsize = fp.seek(0, 2)  # 2 is SEEK_END
        if fsize == 0:
            return
        if fsize < 17:  # 17=4+4+9
            msg = (
                "File size is less than the minimal size "
                "(17 bytes) of Zstandard Seekable Format."
            )
            raise SeekableFormatError(msg)

        # Read footer
        fp.seek(-9, 2)  # 2 is SEEK_END
        footer = fp.read(9)
        frames_number, descriptor, magic_number = self._s_footer.unpack(footer)
        # Check format
        if magic_number != 0x8F92EAB1:
            msg = (
                "The last 4 bytes of the file is not Zstandard Seekable "
                'Format Magic Number (b"\\xb1\\xea\\x92\\x8f)". '
                "SeekableZstdFile class only supports Zstandard Seekable "
                "Format file or 0-size file. To read a zstd file that is "
                "not in Zstandard Seekable Format, use ZstdFile class."
            )
            raise SeekableFormatError(msg)

        # Seek_Table_Descriptor
        self._has_checksum = descriptor & 0b10000000
        if descriptor & 0b01111100:
            msg = (
                f"In Zstandard Seekable Format version {__format_version__}, the "
                "Reserved_Bits in Seek_Table_Descriptor must be 0."
            )
            raise SeekableFormatError(msg)

        # Frame size
        entry_size = 12 if self._has_checksum else 8
        skippable_frame_size = 17 + frames_number * entry_size
        if fsize < skippable_frame_size:
            raise SeekableFormatError(
                "File size is less than expected size of the seek table frame."
            )

        # Read seek table
        fp.seek(-skippable_frame_size, 2)  # 2 is SEEK_END
        skippable_frame = fp.read(skippable_frame_size)
        skippable_magic_number, content_size = self._s_2uint32.unpack_from(
            skippable_frame, 0
        )

        # Check format
        if skippable_magic_number != 0x184D2A5E:
            msg = "Seek table frame's Magic_Number is wrong."
            raise SeekableFormatError(msg)
        if content_size != skippable_frame_size - 8:
            msg = "Seek table frame's Frame_Size is wrong."
            raise SeekableFormatError(msg)

        # No more fp operations
        if seek_to_0:
            fp.seek(0)

        # Parse seek table
        offset = 8
        for idx in range(frames_number):
            if self._has_checksum:
                compressed_size, decompressed_size, _ = self._s_3uint32.unpack_from(
                    skippable_frame, offset
                )
                offset += 12
            else:
                compressed_size, decompressed_size = self._s_2uint32.unpack_from(
                    skippable_frame, offset
                )
                offset += 8

            # Check format
            if compressed_size == 0 and decompressed_size != 0:
                msg = (
                    f"Wrong seek table. The index {idx} frame (0-based) "
                    "is 0 size, but decompressed size is non-zero, "
                    "this is impossible."
                )
                raise SeekableFormatError(msg)

            # Append to seek table
            self.append_entry(compressed_size, decompressed_size)

            # Check format
            if self._full_c_size > fsize - skippable_frame_size:
                msg = (
                    f"Wrong seek table. Since index {idx} frame (0-based), "
                    "the cumulated compressed size is greater than "
                    "file size."
                )
                raise SeekableFormatError(msg)

        # Check format
        if self._full_c_size != fsize - skippable_frame_size:
            raise SeekableFormatError("The cumulated compressed size is wrong")

        # Parsed successfully, save for future use.
        self._seek_frame_size = skippable_frame_size
        self._file_size = fsize

    # Find frame index by decompressed position
    def index_by_dpos(self, pos: int) -> int | None:
        # Array's first item is 0, so need this.
        pos = max(pos, 0)

        i = bisect_right(self._cumulated_d_size, pos)
        if i != self._frames_count + 1:
            return i
        # None means >= EOF
        return None

    def get_frame_sizes(self, i: int) -> tuple[int, int]:
        return (self._cumulated_c_size[i - 1], self._cumulated_d_size[i - 1])

    def get_full_c_size(self) -> int:
        return self._full_c_size

    def get_full_d_size(self) -> int:
        return self._full_d_size

    # Merge the seek table to max_frames frames.
    # The format allows up to 0xFFFF_FFFF frames. When frames
    # number exceeds, use this method to merge.
    def _merge_frames(self, max_frames: int) -> None:
        if self._frames_count <= max_frames:
            return

        # Clear the table
        arr = self._frames
        a, b = divmod(self._frames_count, max_frames)
        self._clear_seek_table()

        # Merge frames
        pos = 0
        for i in range(max_frames):
            # Slice length
            length = (a + (1 if i < b else 0)) * 2

            # Merge
            c_size = 0
            d_size = 0
            for j in range(pos, pos + length, 2):
                c_size += arr[j]
                d_size += arr[j + 1]
            self.append_entry(c_size, d_size)

            pos += length

    def write_seek_table(self, fp: BinaryIO) -> None:
        # Exceeded format limit
        if self._frames_count > 0xFFFFFFFF:
            # Emit a warning
            warnings.warn(
                f"SeekableZstdFile's seek table has {self._frames_count} entries, "
                "which exceeds the maximal value allowed by "
                "Zstandard Seekable Format (0xFFFFFFFF). The "
                "entries will be merged into 0xFFFFFFFF entries, "
                "this may reduce seeking performance.",
                RuntimeWarning,
                3,
            )

            # Merge frames
            self._merge_frames(0xFFFFFFFF)

        # The skippable frame
        offset = 0
        size = 17 + 8 * self._frames_count
        ba = bytearray(size)

        # Header
        self._s_2uint32.pack_into(ba, offset, 0x184D2A5E, size - 8)
        offset += 8
        # Entries
        iter_frames = iter(self._frames)
        for frame_c, frame_d in zip(iter_frames, iter_frames, strict=True):
            self._s_2uint32.pack_into(ba, offset, frame_c, frame_d)
            offset += 8
        # Footer
        self._s_footer.pack_into(ba, offset, self._frames_count, 0, 0x8F92EAB1)

        # Write
        fp.write(ba)

    @property
    def seek_frame_size(self) -> int:
        return self._seek_frame_size

    @property
    def file_size(self) -> int:
        return self._file_size

    def __len__(self) -> int:
        return self._frames_count

    def get_info(self) -> tuple[int, int, int]:
        return (self._frames_count, self._full_c_size, self._full_d_size)


class _EOFSuccess(EOFError):  # noqa: N818
    pass


class _SeekableDecompressReader(io.RawIOBase):
    def __init__(
        self, fp: BinaryIO, zstd_dict: _ZstdDict, option: _Option, read_size: int
    ) -> None:
        # Check fp readable/seekable
        if not hasattr(fp, "readable") or not hasattr(fp, "seekable"):
            raise TypeError(
                "In SeekableZstdFile's reading mode, the file object should "
                "have .readable()/.seekable() methods."
            )
        if not fp.readable():
            raise TypeError(
                "In SeekableZstdFile's reading mode, the file object should "
                "be readable."
            )
        if not fp.seekable():
            raise TypeError(
                "In SeekableZstdFile's reading mode, the file object should "
                "be seekable. If the file object is not seekable, it can be "
                "read sequentially using ZstdFile class."
            )

        self._fp = fp
        self._zstd_dict = zstd_dict
        self._option = option
        self._read_size = read_size

        # Load seek table
        self._seek_table = _SeekTable(read_mode=True)
        self._seek_table.load_seek_table(fp, seek_to_0=True)
        self._size = self._seek_table.get_full_d_size()

        self._pos = 0
        self._decompressor: ZstdDecompressor | None = ZstdDecompressor(
            self._zstd_dict, self._option
        )

    def close(self) -> None:
        self._decompressor = None
        return super().close()

    def readable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True

    def tell(self) -> int:
        return self._pos

    def _decompress(self, size: int) -> bytes:
        """
        Decompress up to size bytes.
        May return b"", in which case try again.
        Raises _EOFSuccess if EOF is reached at frame edge.
        Raises EOFError if EOF is reached elsewhere.
        """
        if self._decompressor is None:  # frame edge
            data = self._fp.read(self._read_size)
            if not data:  # EOF
                raise _EOFSuccess
        elif self._decompressor.needs_input:
            data = self._fp.read(self._read_size)
            if not data:  # EOF
                raise EOFError(
                    "Compressed file ended before the end-of-stream marker was reached"
                )
        else:
            data = self._decompressor.unused_data
            if self._decompressor.eof:  # frame edge
                self._decompressor = None
                if not data:  # may not be at EOF
                    return b""
        if self._decompressor is None:
            self._decompressor = ZstdDecompressor(self._zstd_dict, self._option)
        out = self._decompressor.decompress(data, size)
        self._pos += len(out)
        return out

    def readinto(self, b: Buffer) -> int:
        with memoryview(b) as view, view.cast("B") as byte_view:
            try:
                while True:
                    if out := self._decompress(byte_view.nbytes):
                        byte_view[: len(out)] = out
                        return len(out)
            except _EOFSuccess:
                return 0

    # If the new position is within BufferedReader's buffer,
    # this method may not be called.
    def seek(self, offset: int, whence: int = 0) -> int:
        # offset is absolute file position
        if whence == 0:  # SEEK_SET
            pass
        elif whence == 1:  # SEEK_CUR
            offset = self._pos + offset
        elif whence == 2:  # SEEK_END
            offset = self._size + offset
        else:
            raise ValueError(f"Invalid value for whence: {whence}")

        # Get new frame index
        new_frame = self._seek_table.index_by_dpos(offset)
        # offset >= EOF
        if new_frame is None:
            self._pos = self._size
            self._decompressor = None
            self._fp.seek(self._seek_table.file_size)
            return self._pos

        # Prepare to jump
        old_frame = self._seek_table.index_by_dpos(self._pos)
        c_pos, d_pos = self._seek_table.get_frame_sizes(new_frame)

        # If at P1, seeking to P2 will unnecessarily read the skippable
        # frame. So check self._fp position to skip the skippable frame.
        #       |--data1--|--skippable--|--data2--|
        # cpos:             ^P1
        # dpos:           ^P1             ^P2
        if new_frame == old_frame and offset >= self._pos and self._fp.tell() >= c_pos:
            pass
        else:
            # Jump
            self._pos = d_pos
            self._decompressor = None
            self._fp.seek(c_pos)

        # offset is bytes number to skip forward
        offset -= self._pos
        while offset > 0:
            offset -= len(self._decompress(offset))

        return self._pos

    def get_seek_table_info(self) -> tuple[int, int, int]:
        return self._seek_table.get_info()


# Compared to ZstdFile class, it's important to handle the seekable
# of underlying file object carefully. Need to check seekable in
# each situation. For example, there may be a CD-R file system that
# is seekable when reading, but not seekable when appending.
class SeekableZstdFile(io.BufferedIOBase):
    """This class can only create/write/read Zstandard Seekable Format file,
    or read 0-size file.
    It provides relatively fast seeking ability in read mode.
    """

    # The format uses uint32_t for compressed/decompressed sizes. If flush
    # block a lot, compressed_size may exceed the limit, so set a max size.
    FRAME_MAX_C_SIZE: ClassVar[int] = 2 * 1024 * 1024 * 1024
    # Zstd seekable format's example code also use 1GiB as max content size.
    FRAME_MAX_D_SIZE: ClassVar[int] = 1 * 1024 * 1024 * 1024

    FLUSH_BLOCK: ClassVar[Literal[1]] = ZstdCompressor.FLUSH_BLOCK
    FLUSH_FRAME: ClassVar[Literal[2]] = ZstdCompressor.FLUSH_FRAME

    def __init__(
        self,
        filename: _StrOrBytesPath | BinaryIO,
        mode: Literal["r", "rb", "w", "wb", "a", "ab", "x", "xb"] = "r",
        *,
        level_or_option: _LevelOrOption | _Option = None,
        zstd_dict: _ZstdDict = None,
        read_size: int | _DeprecatedPlaceholder = _DEPRECATED_PLACEHOLDER,  # type: ignore[has-type]
        write_size: int | _DeprecatedPlaceholder = _DEPRECATED_PLACEHOLDER,  # type: ignore[has-type]
        max_frame_content_size: int = 1024 * 1024 * 1024,
    ) -> None:
        """Open a Zstandard Seekable Format file in binary mode. In read mode,
        the file can be 0-size file.

        filename can be either an actual file name (given as a str, bytes, or
        PathLike object), in which case the named file is opened, or it can be
        an existing file object to read from or write to.

        mode can be "r" for reading (default), "w" for (over)writing, "x" for
        creating exclusively, or "a" for appending. These can equivalently be
        given as "rb", "wb", "xb" and "ab" respectively.

        In append mode ("a" or "ab"), filename argument can't be a file object,
        please use file path.

        Parameters
        level_or_option: When it's an int object, it represents compression
            level. When it's a dict object, it contains advanced compression
            parameters. Note, in read mode (decompression), it can only be a
            dict object, that represents decompression option. It doesn't
            support int type compression level in this case.
        zstd_dict: A ZstdDict object, pre-trained dictionary for compression /
            decompression.
        max_frame_content_size: In write/append modes (compression), when
            the uncompressed data size reaches max_frame_content_size, a frame
            is generated automatically. If the size is small, it will increase
            seeking speed, but reduce compression ratio. If the size is large,
            it will reduce seeking speed, but increase compression ratio. You
            can also manually generate a frame using f.flush(f.FLUSH_FRAME).
        """
        if read_size == _DEPRECATED_PLACEHOLDER:
            read_size = 131075
        else:
            warnings.warn(
                "pyzstd.SeekableZstdFile()'s read_size parameter is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
            read_size = cast("int", read_size)
        if write_size == _DEPRECATED_PLACEHOLDER:
            write_size = 131591
        else:
            warnings.warn(
                "pyzstd.SeekableZstdFile()'s write_size parameter is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
            write_size = cast("int", write_size)

        self._fp: BinaryIO | None = None
        self._close_fp = False
        self._mode = _MODE_CLOSED
        self._buffer = None

        if not isinstance(mode, str):
            raise TypeError("mode must be a str")
        mode = mode.removesuffix("b")  # type: ignore[assignment]  # handle rb, wb, xb, ab

        # Read or write mode
        if mode == "r":
            if not isinstance(level_or_option, (type(None), dict)):
                raise TypeError(
                    "In read mode (decompression), level_or_option argument "
                    "should be a dict object, that represents decompression "
                    "option. It doesn't support int type compression level "
                    "in this case."
                )
            if read_size <= 0:
                raise ValueError("read_size argument should > 0")
            if write_size != 131591:
                raise ValueError("write_size argument is only valid in write modes.")
            # Specified max_frame_content_size argument
            if max_frame_content_size != 1024 * 1024 * 1024:
                raise ValueError(
                    "max_frame_content_size argument is only "
                    "valid in write modes (compression)."
                )
            mode_code = _MODE_READ

        elif mode in {"w", "a", "x"}:
            if not isinstance(level_or_option, (type(None), int, dict)):
                raise TypeError(
                    "level_or_option argument should be int or dict object."
                )
            if read_size != 131075:
                raise ValueError("read_size argument is only valid in read mode.")
            if write_size <= 0:
                raise ValueError("write_size argument should > 0")
            if not (0 < max_frame_content_size <= self.FRAME_MAX_D_SIZE):
                raise ValueError(
                    "max_frame_content_size argument should be "
                    f"0 < value <= {self.FRAME_MAX_D_SIZE}, "
                    f"provided value is {max_frame_content_size}."
                )

            # For seekable format
            self._max_frame_content_size = max_frame_content_size
            self._reset_frame_sizes()
            self._seek_table: _SeekTable | None = _SeekTable(read_mode=False)

            mode_code = _MODE_WRITE
            self._compressor: ZstdCompressor | None = ZstdCompressor(
                level_or_option=level_or_option, zstd_dict=zstd_dict
            )
            self._pos = 0

            # Load seek table in append mode
            if mode == "a":
                if not isinstance(filename, (str, bytes, PathLike)):
                    raise TypeError(
                        "In append mode ('a', 'ab'), "
                        "SeekableZstdFile.__init__() method can't "
                        "accept file object as filename argument. "
                        "Please use file path (str/bytes/PathLike)."
                    )

                # Load seek table if file exists
                if isfile(filename):
                    with open(filename, "rb") as f:
                        if not hasattr(f, "seekable") or not f.seekable():
                            raise TypeError(
                                "In SeekableZstdFile's append mode "
                                "('a', 'ab'), the opened 'rb' file "
                                "object should be seekable."
                            )
                        self._seek_table.load_seek_table(f, seek_to_0=False)

        else:
            raise ValueError(f"Invalid mode: {mode!r}")

        # File object
        if isinstance(filename, (str, bytes, PathLike)):
            self._fp = cast("BinaryIO", open(filename, mode + "b"))  # noqa: SIM115
            self._close_fp = True
        elif hasattr(filename, "read") or hasattr(filename, "write"):
            self._fp = filename
        else:
            raise TypeError("filename must be a str, bytes, file or PathLike object")

        self._mode = mode_code

        if self._mode == _MODE_READ:
            raw = _SeekableDecompressReader(
                self._fp,
                zstd_dict=zstd_dict,
                option=cast("_Option", level_or_option),  # checked earlier on
                read_size=read_size,
            )
            self._buffer = io.BufferedReader(raw)

        elif mode == "a":
            if self._fp.seekable():
                self._fp.seek(self._seek_table.get_full_c_size())  # type: ignore[union-attr]
                # Necessary if the current table has many (0, 0) entries
                self._fp.truncate()
            else:
                # Add the seek table frame
                self._seek_table.append_entry(self._seek_table.seek_frame_size, 0)  # type: ignore[union-attr]
                # Emit a warning
                warnings.warn(
                    (
                        "SeekableZstdFile is opened in append mode "
                        "('a', 'ab'), but the underlying file object "
                        "is not seekable. Therefore the seek table (a "
                        "zstd skippable frame) at the end of the file "
                        "can't be overwritten. Each time open such file "
                        "in append mode, it will waste some storage "
                        f"space. {self._seek_table.seek_frame_size} bytes "  # type: ignore[union-attr]
                        "were wasted this time."
                    ),
                    RuntimeWarning,
                    2,
                )

    def _reset_frame_sizes(self) -> None:
        self._current_c_size = 0
        self._current_d_size = 0
        self._left_d_size = self._max_frame_content_size

    def _check_not_closed(self) -> None:
        if self.closed:
            raise ValueError("I/O operation on closed file")

    def _check_can_read(self) -> None:
        if not self.readable():
            raise io.UnsupportedOperation("File not open for reading")

    def _check_can_write(self) -> None:
        if not self.writable():
            raise io.UnsupportedOperation("File not open for writing")

    def close(self) -> None:
        """Flush and close the file.

        May be called more than once without error. Once the file is
        closed, any other operation on it will raise a ValueError.
        """
        if self._mode == _MODE_CLOSED:
            return

        if self._fp is None:
            return
        try:
            if self._mode == _MODE_READ:
                if getattr(self, "_buffer", None):
                    self._buffer.close()  # type: ignore[union-attr]
                    self._buffer = None
            elif self._mode == _MODE_WRITE:
                self.flush(self.FLUSH_FRAME)
                self._seek_table.write_seek_table(self._fp)  # type: ignore[union-attr]
                self._compressor = None
        finally:
            self._mode = _MODE_CLOSED
            self._seek_table = None
            try:
                if self._close_fp:
                    self._fp.close()
            finally:
                self._fp = None
                self._close_fp = False

    def write(self, data: Buffer) -> int:
        """Write a bytes-like object to the file.

        Returns the number of uncompressed bytes written, which is
        always the length of data in bytes. Note that due to buffering,
        the file on disk may not reflect the data written until .flush()
        or .close() is called.
        """
        self._check_can_write()
        # Accept any data that supports the buffer protocol.
        # And memoryview's subview is faster than slice.
        with memoryview(data) as view, view.cast("B") as byte_view:
            nbytes = byte_view.nbytes
            pos = 0

            while nbytes > 0:
                # Write size
                write_size = min(nbytes, self._left_d_size)

                # Compress & write
                compressed = self._compressor.compress(  # type: ignore[union-attr]
                    byte_view[pos : pos + write_size]
                )
                output_size = self._fp.write(compressed)  # type: ignore[union-attr]
                self._pos += write_size

                pos += write_size
                nbytes -= write_size

                # Cumulate
                self._current_c_size += output_size
                self._current_d_size += write_size
                self._left_d_size -= write_size

                # Should flush a frame
                if (
                    self._left_d_size == 0
                    or self._current_c_size >= self.FRAME_MAX_C_SIZE
                ):
                    self.flush(self.FLUSH_FRAME)

            return pos

    def flush(self, mode: Literal[1, 2] = ZstdCompressor.FLUSH_BLOCK) -> None:
        """Flush remaining data to the underlying stream.

        The mode argument can be ZstdFile.FLUSH_BLOCK, ZstdFile.FLUSH_FRAME.
        Abuse of this method will reduce compression ratio, use it only when
        necessary.

        If the program is interrupted afterwards, all data can be recovered.
        To ensure saving to disk, also need to use os.fsync(fd).

        This method does nothing in reading mode.
        """
        if self._mode == _MODE_READ:
            return

        self._check_not_closed()
        if mode not in {self.FLUSH_BLOCK, self.FLUSH_FRAME}:
            raise ValueError(
                "Invalid mode argument, expected either "
                "ZstdFile.FLUSH_FRAME or "
                "ZstdFile.FLUSH_BLOCK"
            )

        if self._compressor.last_mode != mode:  # type: ignore[union-attr]
            # Flush zstd block/frame, and write.
            compressed = self._compressor.flush(mode)  # type: ignore[union-attr]
            output_size = self._fp.write(compressed)  # type: ignore[union-attr]
            if hasattr(self._fp, "flush"):
                self._fp.flush()  # type: ignore[union-attr]

            # Cumulate
            self._current_c_size += output_size
            # self._current_d_size += 0
            # self._left_d_size -= 0

        if mode == self.FLUSH_FRAME and self._current_c_size != 0:
            # Add an entry to seek table
            self._seek_table.append_entry(self._current_c_size, self._current_d_size)  # type: ignore[union-attr]
            self._reset_frame_sizes()

    def read(self, size: int | None = -1) -> bytes:
        """Read up to size uncompressed bytes from the file.

        If size is negative or omitted, read until EOF is reached.
        Returns b"" if the file is already at EOF.
        """
        if size is None:
            size = -1
        self._check_can_read()
        return self._buffer.read(size)  # type: ignore[union-attr]

    def read1(self, size: int = -1) -> bytes:
        """Read up to size uncompressed bytes, while trying to avoid
        making multiple reads from the underlying stream. Reads up to a
        buffer's worth of data if size is negative.

        Returns b"" if the file is at EOF.
        """
        self._check_can_read()
        if size < 0:
            size = io.DEFAULT_BUFFER_SIZE
        return self._buffer.read1(size)  # type: ignore[union-attr]

    def readinto(self, b: Buffer) -> int:
        """Read bytes into b.

        Returns the number of bytes read (0 for EOF).
        """
        self._check_can_read()
        return self._buffer.readinto(b)  # type: ignore[union-attr]

    def readinto1(self, b: Buffer) -> int:
        """Read bytes into b, while trying to avoid making multiple reads
        from the underlying stream.

        Returns the number of bytes read (0 for EOF).
        """
        self._check_can_read()
        return self._buffer.readinto1(b)  # type: ignore[union-attr]

    def readline(self, size: int | None = -1) -> bytes:
        """Read a line of uncompressed bytes from the file.

        The terminating newline (if present) is retained. If size is
        non-negative, no more than size bytes will be read (in which
        case the line may be incomplete). Returns b'' if already at EOF.
        """
        self._check_can_read()
        return self._buffer.readline(size)  # type: ignore[union-attr]

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """Change the file position.

        The new position is specified by offset, relative to the
        position indicated by whence. Possible values for whence are:

            0: start of stream (default): offset must not be negative
            1: current stream position
            2: end of stream; offset must not be positive

        Returns the new file position.

        Note that seeking is emulated, so depending on the arguments,
        this operation may be extremely slow.
        """
        self._check_can_read()
        return self._buffer.seek(offset, whence)  # type: ignore[union-attr]

    def peek(self, size: int = -1) -> bytes:
        """Return buffered data without advancing the file position.

        Always returns at least one byte of data, unless at EOF.
        The exact number of bytes returned is unspecified.
        """
        self._check_can_read()
        return self._buffer.peek(size)  # type: ignore[union-attr]

    def __iter__(self) -> Self:
        self._check_can_read()
        return self

    def __next__(self) -> bytes:
        self._check_can_read()
        if ret := self._buffer.readline():  # type: ignore[union-attr]
            return ret
        raise StopIteration

    def tell(self) -> int:
        """Return the current file position."""
        self._check_not_closed()
        if self._mode == _MODE_READ:
            return self._buffer.tell()  # type: ignore[union-attr]
        if self._mode == _MODE_WRITE:
            return self._pos
        raise RuntimeError  # impossible code path

    def fileno(self) -> int:
        """Return the file descriptor for the underlying file."""
        self._check_not_closed()
        return self._fp.fileno()  # type: ignore[union-attr]

    @property
    def name(self) -> str:
        """Return the file name for the underlying file."""
        self._check_not_closed()
        return self._fp.name  # type: ignore[union-attr]

    @property
    def closed(self) -> bool:
        """True if this file is closed."""
        return self._mode == _MODE_CLOSED

    def writable(self) -> bool:
        """Return whether the file was opened for writing."""
        self._check_not_closed()
        return self._mode == _MODE_WRITE

    def readable(self) -> bool:
        """Return whether the file was opened for reading."""
        self._check_not_closed()
        return self._mode == _MODE_READ

    def seekable(self) -> bool:
        """Return whether the file supports seeking."""
        return self.readable() and self._buffer.seekable()  # type: ignore[union-attr]

    @property
    def seek_table_info(self) -> tuple[int, int, int] | None:
        """A tuple: (frames_number, compressed_size, decompressed_size)
        1, Frames_number and compressed_size don't count the seek table
           frame (a zstd skippable frame at the end of the file).
        2, In write modes, the part of data that has not been flushed to
           frames is not counted.
        3, If the SeekableZstdFile object is closed, it's None.
        """
        if self._mode == _MODE_WRITE:
            return self._seek_table.get_info()  # type: ignore[union-attr]
        if self._mode == _MODE_READ:
            return self._buffer.raw.get_seek_table_info()  # type: ignore[union-attr]
        return None

    @staticmethod
    def is_seekable_format_file(filename: _StrOrBytesPath | BinaryIO) -> bool:
        """Check if a file is Zstandard Seekable Format file or 0-size file.

        It parses the seek table at the end of the file, returns True if no
        format error.

        filename can be either a file path (str/bytes/PathLike), or can be an
        existing file object in reading mode.
        """
        # Check argument
        if isinstance(filename, (str, bytes, PathLike)):
            fp: BinaryIO = open(filename, "rb")  # noqa: SIM115
            is_file_path = True
        elif (
            hasattr(filename, "readable")
            and filename.readable()
            and hasattr(filename, "seekable")
            and filename.seekable()
        ):
            fp = filename
            is_file_path = False
            orig_pos = fp.tell()
        else:
            raise TypeError(
                "filename argument should be a str/bytes/PathLike object, "
                "or a file object that is readable and seekable."
            )

        # Write mode uses less RAM
        table = _SeekTable(read_mode=False)
        try:
            # Read/Parse the seek table
            table.load_seek_table(fp, seek_to_0=False)
        except SeekableFormatError:
            ret = False
        else:
            ret = True
        finally:
            if is_file_path:
                fp.close()
            else:
                fp.seek(orig_pos)

        return ret



from collections.abc import Callable, Mapping
from enum import IntEnum
from io import TextIOWrapper
from os import PathLike
import sys
from typing import (
    BinaryIO,
    ClassVar,
    Literal,
    NamedTuple,
    NoReturn,
    TypeAlias,
    cast,
    overload,
)
import warnings

if sys.version_info < (3, 14):
    from backports import zstd
else:
    from compression import zstd

if sys.version_info < (3, 13):
    from typing_extensions import deprecated
else:
    from warnings import deprecated

if sys.version_info < (3, 12):
    from typing_extensions import Buffer
else:
    from collections.abc import Buffer

from pyzstd._version import __version__  # noqa: F401

__doc__ = """\
Python bindings to Zstandard (zstd) compression library, the API style is
similar to Python's bz2/lzma/zlib modules.

Command line interface of this module: python -m pyzstd --help

Documentation: https://pyzstd.readthedocs.io
GitHub: https://github.com/Rogdham/pyzstd
PyPI: https://pypi.org/project/pyzstd"""

__all__ = (
    "CParameter",
    "DParameter",
    "EndlessZstdDecompressor",
    "RichMemZstdCompressor",
    "SeekableFormatError",
    "SeekableZstdFile",
    "Strategy",
    "ZstdCompressor",
    "ZstdDecompressor",
    "ZstdDict",
    "ZstdError",
    "ZstdFile",
    "compress",
    "compress_stream",
    "compressionLevel_values",
    "decompress",
    "decompress_stream",
    "finalize_dict",
    "get_frame_info",
    "get_frame_size",
    "open",
    "richmem_compress",
    "train_dict",
    "zstd_support_multithread",
    "zstd_version",
    "zstd_version_info",
)


class _DeprecatedPlaceholder:
    def __repr__(self) -> str:
        return "<DEPRECATED>"


_DEPRECATED_PLACEHOLDER = _DeprecatedPlaceholder()


Strategy = zstd.Strategy
ZstdError = zstd.ZstdError
ZstdDict = zstd.ZstdDict
train_dict = zstd.train_dict
finalize_dict = zstd.finalize_dict
get_frame_info = zstd.get_frame_info
get_frame_size = zstd.get_frame_size
zstd_version = zstd.zstd_version
zstd_version_info = zstd.zstd_version_info


class CParameter(IntEnum):
    """Compression parameters"""

    compressionLevel = zstd.CompressionParameter.compression_level  # noqa: N815
    windowLog = zstd.CompressionParameter.window_log  # noqa: N815
    hashLog = zstd.CompressionParameter.hash_log  # noqa: N815
    chainLog = zstd.CompressionParameter.chain_log  # noqa: N815
    searchLog = zstd.CompressionParameter.search_log  # noqa: N815
    minMatch = zstd.CompressionParameter.min_match  # noqa: N815
    targetLength = zstd.CompressionParameter.target_length  # noqa: N815
    strategy = zstd.CompressionParameter.strategy
    targetCBlockSize = 130  # not part of PEP-784 # noqa: N815

    enableLongDistanceMatching = zstd.CompressionParameter.enable_long_distance_matching  # noqa: N815
    ldmHashLog = zstd.CompressionParameter.ldm_hash_log  # noqa: N815
    ldmMinMatch = zstd.CompressionParameter.ldm_min_match  # noqa: N815
    ldmBucketSizeLog = zstd.CompressionParameter.ldm_bucket_size_log  # noqa: N815
    ldmHashRateLog = zstd.CompressionParameter.ldm_hash_rate_log  # noqa: N815

    contentSizeFlag = zstd.CompressionParameter.content_size_flag  # noqa: N815
    checksumFlag = zstd.CompressionParameter.checksum_flag  # noqa: N815
    dictIDFlag = zstd.CompressionParameter.dict_id_flag  # noqa: N815

    nbWorkers = zstd.CompressionParameter.nb_workers  # noqa: N815
    jobSize = zstd.CompressionParameter.job_size  # noqa: N815
    overlapLog = zstd.CompressionParameter.overlap_log  # noqa: N815

    def bounds(self) -> tuple[int, int]:
        """Return lower and upper bounds of a compression parameter, both inclusive."""
        return zstd.CompressionParameter(self).bounds()


class DParameter(IntEnum):
    """Decompression parameters"""

    windowLogMax = zstd.DecompressionParameter.window_log_max  # noqa: N815

    def bounds(self) -> tuple[int, int]:
        """Return lower and upper bounds of a decompression parameter, both inclusive."""
        return zstd.DecompressionParameter(self).bounds()


_LevelOrOption: TypeAlias = int | Mapping[int, int] | None
_Option: TypeAlias = Mapping[int, int] | None
_ZstdDict: TypeAlias = ZstdDict | tuple[ZstdDict, int] | None
_StrOrBytesPath: TypeAlias = str | bytes | PathLike[str] | PathLike[bytes]


def _convert_level_or_option(
    level_or_option: _LevelOrOption | _Option, mode: str
) -> Mapping[int, int] | None:
    """Transform pyzstd params into PEP-784 `options` param"""
    if not isinstance(mode, str):
        raise TypeError(f"Invalid mode type: {mode}")
    read_mode = mode.startswith("r")
    if isinstance(level_or_option, int):
        if read_mode:
            raise TypeError(
                "In read mode (decompression), level_or_option argument "
                "should be a dict object, that represents decompression "
                "option. It doesn't support int type compression level "
                "in this case."
            )
        return {
            CParameter.compressionLevel: level_or_option,
        }
    if level_or_option is not None:
        invalid_class = CParameter if read_mode else DParameter
        for key in level_or_option:
            if isinstance(key, invalid_class):
                raise TypeError(
                    "Key of compression option dict should "
                    f"NOT be {invalid_class.__name__}."
                )
    return level_or_option


class ZstdCompressor:
    """A streaming compressor. Thread-safe at method level."""

    CONTINUE: ClassVar[Literal[0]] = zstd.ZstdCompressor.CONTINUE
    """Used for mode parameter in .compress() method.

    Collect more data, encoder decides when to output compressed result, for optimal
    compression ratio. Usually used for traditional streaming compression.
    """

    FLUSH_BLOCK: ClassVar[Literal[1]] = zstd.ZstdCompressor.FLUSH_BLOCK
    """Used for mode parameter in .compress(), .flush() methods.

    Flush any remaining data, but don't close the current frame. Usually used for
    communication scenarios.

    If there is data, it creates at least one new block, that can be decoded
    immediately on reception. If no remaining data, no block is created, return b''.

    Note: Abuse of this mode will reduce compression ratio. Use it only when
    necessary.
    """

    FLUSH_FRAME: ClassVar[Literal[2]] = zstd.ZstdCompressor.FLUSH_FRAME
    """Used for mode parameter in .compress(), .flush() methods.

    Flush any remaining data, and close the current frame. Usually used for
    traditional flush.

    Since zstd data consists of one or more independent frames, data can still be
    provided after a frame is closed.

    Note: Abuse of this mode will reduce compression ratio, and some programs can
    only decompress single frame data. Use it only when necessary.
    """

    def __init__(
        self, level_or_option: _LevelOrOption = None, zstd_dict: _ZstdDict = None
    ) -> None:
        """Initialize a ZstdCompressor object.

        Parameters
        level_or_option: When it's an int object, it represents the compression level.
                         When it's a dict object, it contains advanced compression
                         parameters.
        zstd_dict:       A ZstdDict object, pre-trained zstd dictionary.
        """
        zstd_dict = cast(
            "ZstdDict | None", zstd_dict
        )  # https://github.com/python/typeshed/pull/15113
        self._compressor = zstd.ZstdCompressor(
            options=_convert_level_or_option(level_or_option, "w"), zstd_dict=zstd_dict
        )

    def compress(
        self, data: Buffer, mode: Literal[0, 1, 2] = zstd.ZstdCompressor.CONTINUE
    ) -> bytes:
        """Provide data to the compressor object.
        Return a chunk of compressed data if possible, or b'' otherwise.

        Parameters
        data: A bytes-like object, data to be compressed.
        mode: Can be these 3 values .CONTINUE, .FLUSH_BLOCK, .FLUSH_FRAME.
        """
        return self._compressor.compress(data, mode)

    def flush(self, mode: Literal[1, 2] = zstd.ZstdCompressor.FLUSH_FRAME) -> bytes:
        """Flush any remaining data in internal buffer.

        Since zstd data consists of one or more independent frames, the compressor
        object can still be used after this method is called.

        Parameter
        mode: Can be these 2 values .FLUSH_FRAME, .FLUSH_BLOCK.
        """
        return self._compressor.flush(mode)

    def _set_pledged_input_size(self, size: int | None) -> None:
        """*This is an undocumented method, because it may be used incorrectly.*

        Set uncompressed content size of a frame, the size will be written into the
        frame header.
        1, If called when (.last_mode != .FLUSH_FRAME), a RuntimeError will be raised.
        2, If the actual size doesn't match the value, a ZstdError will be raised, and
           the last compressed chunk is likely to be lost.
        3, The size is only valid for one frame, then it restores to "unknown size".

        Parameter
        size: Uncompressed content size of a frame, None means "unknown size".
        """
        return self._compressor.set_pledged_input_size(size)

    @property
    def last_mode(self) -> Literal[0, 1, 2]:
        """The last mode used to this compressor object, its value can be .CONTINUE,
        .FLUSH_BLOCK, .FLUSH_FRAME. Initialized to .FLUSH_FRAME.

        It can be used to get the current state of a compressor, such as, data flushed,
        a frame ended.
        """
        return self._compressor.last_mode

    def __reduce__(self) -> NoReturn:
        raise TypeError(f"Cannot pickle {type(self)} object.")


class ZstdDecompressor:
    """A streaming decompressor, it stops after a frame is decompressed.
    Thread-safe at method level."""

    def __init__(self, zstd_dict: _ZstdDict = None, option: _Option = None) -> None:
        """Initialize a ZstdDecompressor object.

        Parameters
        zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
        option:    A dict object that contains advanced decompression parameters.
        """
        zstd_dict = cast(
            "ZstdDict | None", zstd_dict
        )  # https://github.com/python/typeshed/pull/15113
        self._decompressor = zstd.ZstdDecompressor(
            zstd_dict=zstd_dict, options=_convert_level_or_option(option, "r")
        )

    def decompress(self, data: Buffer, max_length: int = -1) -> bytes:
        """Decompress data, return a chunk of decompressed data if possible, or b''
        otherwise.

        It stops after a frame is decompressed.

        Parameters
        data:       A bytes-like object, zstd data to be decompressed.
        max_length: Maximum size of returned data. When it is negative, the size of
                    output buffer is unlimited. When it is nonnegative, returns at
                    most max_length bytes of decompressed data.
        """
        return self._decompressor.decompress(data, max_length)

    @property
    def eof(self) -> bool:
        """True means the end of the first frame has been reached. If decompress data
        after that, an EOFError exception will be raised."""
        return self._decompressor.eof

    @property
    def needs_input(self) -> bool:
        """If the max_length output limit in .decompress() method has been reached, and
        the decompressor has (or may has) unconsumed input data, it will be set to
        False. In this case, pass b'' to .decompress() method may output further data.
        """
        return self._decompressor.needs_input

    @property
    def unused_data(self) -> bytes:
        """A bytes object. When ZstdDecompressor object stops after a frame is
        decompressed, unused input data after the frame. Otherwise this will be b''."""
        return self._decompressor.unused_data

    def __reduce__(self) -> NoReturn:
        raise TypeError(f"Cannot pickle {type(self)} object.")


class EndlessZstdDecompressor:
    """A streaming decompressor, accepts multiple concatenated frames.
    Thread-safe at method level."""

    def __init__(self, zstd_dict: _ZstdDict = None, option: _Option = None) -> None:
        """Initialize an EndlessZstdDecompressor object.

        Parameters
        zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
        option:    A dict object that contains advanced decompression parameters.
        """
        self._zstd_dict = cast(
            "ZstdDict | None", zstd_dict
        )  # https://github.com/python/typeshed/pull/15113
        self._options = _convert_level_or_option(option, "r")
        self._reset()

    def _reset(self, data: bytes = b"") -> None:
        self._decompressor = zstd.ZstdDecompressor(
            zstd_dict=self._zstd_dict, options=self._options
        )
        self._buffer = data
        self._at_frame_edge = not data

    def decompress(self, data: Buffer, max_length: int = -1) -> bytes:
        """Decompress data, return a chunk of decompressed data if possible, or b''
        otherwise.

        Parameters
        data:       A bytes-like object, zstd data to be decompressed.
        max_length: Maximum size of returned data. When it is negative, the size of
                    output buffer is unlimited. When it is nonnegative, returns at
                    most max_length bytes of decompressed data.
        """
        if not isinstance(data, bytes) or not isinstance(max_length, int):
            raise TypeError
        self._buffer += data
        self._at_frame_edge &= not self._buffer
        out = b""
        while True:
            try:
                out += self._decompressor.decompress(self._buffer, max_length)
            except ZstdError:
                self._reset()
                raise
            if self._decompressor.eof:
                self._reset(self._decompressor.unused_data)
                max_length -= len(out)
            else:
                self._buffer = b""
                break
        return out

    @property
    def at_frame_edge(self) -> bool:
        """True when both the input and output streams are at a frame edge, means a
        frame is completely decoded and fully flushed, or the decompressor just be
        initialized.

        This flag could be used to check data integrity in some cases.
        """
        return self._at_frame_edge

    @property
    def needs_input(self) -> bool:
        """If the max_length output limit in .decompress() method has been reached, and
        the decompressor has (or may has) unconsumed input data, it will be set to
        False. In this case, pass b'' to .decompress() method may output further data.
        """
        return not self._buffer and (
            self._at_frame_edge or self._decompressor.needs_input
        )

    def __reduce__(self) -> NoReturn:
        raise TypeError(f"Cannot pickle {type(self)} object.")


def compress(
    data: Buffer, level_or_option: _LevelOrOption = None, zstd_dict: _ZstdDict = None
) -> bytes:
    """Compress a block of data, return a bytes object.

    Compressing b'' will get an empty content frame (9 bytes or more).

    Parameters
    data:            A bytes-like object, data to be compressed.
    level_or_option: When it's an int object, it represents compression level.
                     When it's a dict object, it contains advanced compression
                     parameters.
    zstd_dict:       A ZstdDict object, pre-trained dictionary for compression.
    """
    zstd_dict = cast(
        "ZstdDict | None", zstd_dict
    )  # https://github.com/python/typeshed/pull/15113
    return zstd.compress(
        data,
        options=_convert_level_or_option(level_or_option, "w"),
        zstd_dict=zstd_dict,
    )


def decompress(
    data: Buffer, zstd_dict: _ZstdDict = None, option: _Option = None
) -> bytes:
    """Decompress a zstd data, return a bytes object.

    Support multiple concatenated frames.

    Parameters
    data:      A bytes-like object, compressed zstd data.
    zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
    option:    A dict object, contains advanced decompression parameters.
    """
    zstd_dict = cast(
        "ZstdDict | None", zstd_dict
    )  # https://github.com/python/typeshed/pull/15113
    return zstd.decompress(
        data, options=_convert_level_or_option(option, "r"), zstd_dict=zstd_dict
    )


@deprecated(
    "See https://pyzstd.readthedocs.io/en/stable/deprecated.html for alternatives to pyzstd.RichMemZstdCompressor"
)
class RichMemZstdCompressor:
    def __init__(
        self, level_or_option: _LevelOrOption = None, zstd_dict: _ZstdDict = None
    ) -> None:
        self._options = _convert_level_or_option(level_or_option, "w")
        self._zstd_dict = cast(
            "ZstdDict | None", zstd_dict
        )  # https://github.com/python/typeshed/pull/15113

    def compress(self, data: Buffer) -> bytes:
        return zstd.compress(data, options=self._options, zstd_dict=self._zstd_dict)

    def __reduce__(self) -> NoReturn:
        raise TypeError(f"Cannot pickle {type(self)} object.")


richmem_compress = deprecated(
    "See https://pyzstd.readthedocs.io/en/stable/deprecated.html for alternatives to pyzstd.richmem_compress"
)(compress)


class ZstdFile(zstd.ZstdFile):
    """A file object providing transparent zstd (de)compression.

    A ZstdFile can act as a wrapper for an existing file object, or refer
    directly to a named file on disk.

    Note that ZstdFile provides a *binary* file interface - data read is
    returned as bytes, and data to be written should be an object that
    supports the Buffer Protocol.
    """

    def __init__(
        self,
        filename: _StrOrBytesPath | BinaryIO,
        mode: Literal["r", "rb", "w", "wb", "x", "xb", "a", "ab"] = "r",
        *,
        level_or_option: _LevelOrOption | _Option = None,
        zstd_dict: _ZstdDict = None,
        read_size: int | _DeprecatedPlaceholder = _DEPRECATED_PLACEHOLDER,
        write_size: int | _DeprecatedPlaceholder = _DEPRECATED_PLACEHOLDER,
    ) -> None:
        """Open a zstd compressed file in binary mode.

        filename can be either an actual file name (given as a str, bytes, or
        PathLike object), in which case the named file is opened, or it can be
        an existing file object to read from or write to.

        mode can be "r" for reading (default), "w" for (over)writing, "x" for
        creating exclusively, or "a" for appending. These can equivalently be
        given as "rb", "wb", "xb" and "ab" respectively.

        Parameters
        level_or_option: When it's an int object, it represents compression
            level. When it's a dict object, it contains advanced compression
            parameters. Note, in read mode (decompression), it can only be a
            dict object, that represents decompression option. It doesn't
            support int type compression level in this case.
        zstd_dict: A ZstdDict object, pre-trained dictionary for compression /
            decompression.
        """
        if read_size != _DEPRECATED_PLACEHOLDER:
            warnings.warn(
                "pyzstd.ZstdFile()'s read_size parameter is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
        if write_size != _DEPRECATED_PLACEHOLDER:
            warnings.warn(
                "pyzstd.ZstdFile()'s write_size parameter is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
        zstd_dict = cast(
            "ZstdDict | None", zstd_dict
        )  # https://github.com/python/typeshed/pull/15113
        super().__init__(
            filename,
            mode,
            options=_convert_level_or_option(level_or_option, mode),
            zstd_dict=zstd_dict,
        )


@overload
def open(  # noqa: A001
    filename: _StrOrBytesPath | BinaryIO,
    mode: Literal["r", "rb", "w", "wb", "a", "ab", "x", "xb"] = "rb",
    *,
    level_or_option: _LevelOrOption | _Option = None,
    zstd_dict: _ZstdDict = None,
    encoding: None = None,
    errors: None = None,
    newline: None = None,
) -> zstd.ZstdFile: ...


@overload
def open(  # noqa: A001
    filename: _StrOrBytesPath | BinaryIO,
    mode: Literal["rt", "wt", "at", "xt"],
    *,
    level_or_option: _LevelOrOption | _Option = None,
    zstd_dict: _ZstdDict = None,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
) -> TextIOWrapper: ...


def open(  # noqa: A001
    filename: _StrOrBytesPath | BinaryIO,
    mode: Literal[
        "r", "rb", "w", "wb", "a", "ab", "x", "xb", "rt", "wt", "at", "xt"
    ] = "rb",
    *,
    level_or_option: _LevelOrOption | _Option = None,
    zstd_dict: _ZstdDict = None,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
) -> zstd.ZstdFile | TextIOWrapper:
    """Open a zstd compressed file in binary or text mode.

    filename can be either an actual file name (given as a str, bytes, or
    PathLike object), in which case the named file is opened, or it can be an
    existing file object to read from or write to.

    The mode parameter can be "r", "rb" (default), "w", "wb", "x", "xb", "a",
    "ab" for binary mode, or "rt", "wt", "xt", "at" for text mode.

    The level_or_option and zstd_dict parameters specify the settings, as for
    ZstdCompressor, ZstdDecompressor and ZstdFile.

    When using read mode (decompression), the level_or_option parameter can
    only be a dict object, that represents decompression option. It doesn't
    support int type compression level in this case.

    For binary mode, this function is equivalent to the ZstdFile constructor:
    ZstdFile(filename, mode, ...). In this case, the encoding, errors and
    newline parameters must not be provided.

    For text mode, an ZstdFile object is created, and wrapped in an
    io.TextIOWrapper instance with the specified encoding, error handling
    behavior, and line ending(s).
    """
    zstd_dict = cast(
        "ZstdDict | None", zstd_dict
    )  # https://github.com/python/typeshed/pull/15113
    return zstd.open(
        filename,
        mode,
        options=_convert_level_or_option(level_or_option, mode),
        zstd_dict=zstd_dict,
        encoding=encoding,
        errors=errors,
        newline=newline,
    )


def _create_callback(
    output_stream: BinaryIO | None,
    callback: Callable[[int, int, memoryview, memoryview], None] | None,
) -> Callable[[int, int, bytes, bytes], None]:
    if output_stream is None:
        if callback is None:
            raise TypeError(
                "At least one of output_stream argument and callback argument should be non-None."
            )

        def cb(
            total_input: int, total_output: int, data_in: bytes, data_out: bytes
        ) -> None:
            callback(
                total_input, total_output, memoryview(data_in), memoryview(data_out)
            )

    elif callback is None:

        def cb(
            total_input: int,  # noqa: ARG001
            total_output: int,  # noqa: ARG001
            data_in: bytes,  # noqa: ARG001
            data_out: bytes,
        ) -> None:
            output_stream.write(data_out)

    else:

        def cb(
            total_input: int, total_output: int, data_in: bytes, data_out: bytes
        ) -> None:
            output_stream.write(data_out)
            callback(
                total_input, total_output, memoryview(data_in), memoryview(data_out)
            )

    return cb


@deprecated(
    "See https://pyzstd.readthedocs.io/en/stable/deprecated.html for alternatives to pyzstd.compress_stream"
)
def compress_stream(
    input_stream: BinaryIO,
    output_stream: BinaryIO | None,
    *,
    level_or_option: _LevelOrOption = None,
    zstd_dict: _ZstdDict = None,
    pledged_input_size: int | None = None,
    read_size: int = 131_072,
    write_size: int | _DeprecatedPlaceholder = _DEPRECATED_PLACEHOLDER,  # noqa: ARG001
    callback: Callable[[int, int, memoryview, memoryview], None] | None = None,
) -> tuple[int, int]:
    """Compresses input_stream and writes the compressed data to output_stream, it
    doesn't close the streams.

    ----
    DEPRECATION NOTICE
    The (de)compress_stream are deprecated and will be removed in a future version.
    See https://pyzstd.readthedocs.io/en/stable/deprecated.html for alternatives
    ----

    If input stream is b'', nothing will be written to output stream.

    Return a tuple, (total_input, total_output), the items are int objects.

    Parameters
    input_stream: Input stream that has a .readinto(b) method.
    output_stream: Output stream that has a .write(b) method. If use callback
        function, this parameter can be None.
    level_or_option: When it's an int object, it represents the compression
        level. When it's a dict object, it contains advanced compression
        parameters.
    zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
    pledged_input_size: If set this parameter to the size of input data, the
        size will be written into the frame header. If the actual input data
        doesn't match it, a ZstdError will be raised.
    read_size: Input buffer size, in bytes.
    callback: A callback function that accepts four parameters:
        (total_input, total_output, read_data, write_data), the first two are
        int objects, the last two are readonly memoryview objects.
    """
    if not hasattr(input_stream, "read"):
        raise TypeError("input_stream argument should have a .read() method.")
    if output_stream is not None and not hasattr(output_stream, "write"):
        raise TypeError("output_stream argument should have a .write() method.")
    if read_size < 1:
        raise ValueError("read_size argument should be a positive number.")
    callback = _create_callback(output_stream, callback)
    total_input = 0
    total_output = 0
    compressor = ZstdCompressor(level_or_option, zstd_dict)
    if pledged_input_size is not None and pledged_input_size != 2**64 - 1:
        compressor._set_pledged_input_size(pledged_input_size)  # noqa: SLF001
    while data_in := input_stream.read(read_size):
        total_input += len(data_in)
        data_out = compressor.compress(data_in)
        total_output += len(data_out)
        callback(total_input, total_output, data_in, data_out)
    if not total_input:
        return total_input, total_output
    data_out = compressor.flush()
    total_output += len(data_out)
    callback(total_input, total_output, b"", data_out)
    return total_input, total_output


@deprecated(
    "See https://pyzstd.readthedocs.io/en/stable/deprecated.html for alternatives to pyzstd.decompress_stream"
)
def decompress_stream(
    input_stream: BinaryIO,
    output_stream: BinaryIO | None,
    *,
    zstd_dict: _ZstdDict = None,
    option: _Option = None,
    read_size: int = 131_075,
    write_size: int = 131_072,
    callback: Callable[[int, int, memoryview, memoryview], None] | None = None,
) -> tuple[int, int]:
    """Decompresses input_stream and writes the decompressed data to output_stream,
    it doesn't close the streams.

    ----
    DEPRECATION NOTICE
    The (de)compress_stream are deprecated and will be removed in a future version.
    See https://pyzstd.readthedocs.io/en/stable/deprecated.html for alternatives
    ----

    Supports multiple concatenated frames.

    Return a tuple, (total_input, total_output), the items are int objects.

    Parameters
    input_stream: Input stream that has a .readinto(b) method.
    output_stream: Output stream that has a .write(b) method. If use callback
        function, this parameter can be None.
    zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
    option: A dict object, contains advanced decompression parameters.
    read_size: Input buffer size, in bytes.
    write_size: Output buffer size, in bytes.
    callback: A callback function that accepts four parameters:
        (total_input, total_output, read_data, write_data), the first two are
        int objects, the last two are readonly memoryview objects.
    """
    if not hasattr(input_stream, "read"):
        raise TypeError("input_stream argument should have a .read() method.")
    if output_stream is not None and not hasattr(output_stream, "write"):
        raise TypeError("output_stream argument should have a .write() method.")
    if read_size < 1 or write_size < 1:
        raise ValueError(
            "read_size argument and write_size argument should be positive numbers."
        )
    callback = _create_callback(output_stream, callback)
    total_input = 0
    total_output = 0
    decompressor = EndlessZstdDecompressor(zstd_dict, option)
    while True:
        if decompressor.needs_input:
            data_in = input_stream.read(read_size)
            if not data_in:
                break
        else:
            data_in = b""
        total_input += len(data_in)
        data_out = decompressor.decompress(data_in, write_size)
        total_output += len(data_out)
        callback(total_input, total_output, data_in, data_out)
    if not decompressor.at_frame_edge:
        raise ZstdError(
            "Decompression failed: zstd data ends in an incomplete frame,"
            " maybe the input data was truncated."
            f" Total input {total_input} bytes, total output {total_output} bytes."
        )
    return total_input, total_output


class CompressionValues(NamedTuple):
    default: int
    min: int
    max: int


compressionLevel_values = CompressionValues(  # noqa: N816
    zstd.COMPRESSION_LEVEL_DEFAULT, *CParameter.compressionLevel.bounds()
)
zstd_support_multithread = CParameter.nbWorkers.bounds() != (0, 0)


# import here to avoid circular dependency issues
from ._seekable_zstdfile import SeekableFormatError, SeekableZstdFile  # noqa: E402



# CLI of pyzstd module: python -m pyzstd --help
import argparse
from collections.abc import Mapping, Sequence
import os
from shutil import copyfileobj
from time import time
from typing import Any, BinaryIO, Protocol, cast

from pyzstd import (
    CParameter,
    DParameter,
    ZstdDict,
    ZstdFile,
    compressionLevel_values,
    train_dict,
    zstd_version,
)
from pyzstd import __version__ as pyzstd_version


class Args(Protocol):
    dict: str
    f: bool
    compress: str
    tar_input_dir: str
    level: int
    threads: int
    long: int
    checksum: bool
    write_dictID: bool  # noqa: N815
    decompress: str
    tar_output_dir: str
    test: str | None
    windowLogMax: int  # noqa: N815
    train: str
    maxdict: int
    dictID: int  # noqa: N815
    output: BinaryIO | None
    input: BinaryIO | None
    zd: ZstdDict | None


# buffer sizes recommended by zstd
C_READ_BUFFER = 131072
D_READ_BUFFER = 131075


# open output file and assign to args.output
def open_output(args: Args, path: str) -> None:
    if not args.f and os.path.isfile(path):
        answer = input(f"output file already exists:\n{path}\noverwrite? (y/n) ")
        print()
        if answer != "y":
            import sys

            sys.exit()
    args.output = open(path, "wb")  # noqa: SIM115


def close_files(args: Args) -> None:
    if args.input is not None:
        args.input.close()

    if args.output is not None:
        args.output.close()


def compress_option(args: Args) -> Mapping[int, int]:
    # threads message
    if args.threads == 0:
        threads_msg = "single-thread mode"
    else:
        threads_msg = f"multi-thread mode, {args.threads} threads."

    # long mode
    if args.long >= 0:
        use_long = 1
        window_log = args.long
        long_msg = f"yes, windowLog is {window_log}."
    else:
        use_long = 0
        window_log = 0
        long_msg = "no"

    # option
    option: Mapping[int, int] = {
        CParameter.compressionLevel: args.level,
        CParameter.nbWorkers: args.threads,
        CParameter.enableLongDistanceMatching: use_long,
        CParameter.windowLog: window_log,
        CParameter.checksumFlag: args.checksum,
        CParameter.dictIDFlag: args.write_dictID,
    }

    # pre-compress message
    msg = (
        f" - compression level: {args.level}\n"
        f" - threads: {threads_msg}\n"
        f" - long mode: {long_msg}\n"
        f" - zstd dictionary: {args.zd}\n"
        f" - add checksum: {args.checksum}"
    )
    print(msg)

    return option


def compress(args: Args) -> None:
    args.input = cast("BinaryIO", args.input)

    # output file
    if args.output is None:
        open_output(args, args.input.name + ".zst")
        args.output = cast("BinaryIO", args.output)

    # pre-compress message
    msg = (
        "Compress file:\n"
        f" - input file : {args.input.name}\n"
        f" - output file: {args.output.name}"
    )
    print(msg)

    # option
    option = compress_option(args)

    # compress
    t1 = time()
    with ZstdFile(args.output, "w", level_or_option=option, zstd_dict=args.zd) as fout:
        copyfileobj(args.input, fout)
    t2 = time()
    in_size = args.input.tell()
    out_size = args.output.tell()
    close_files(args)

    # post-compress message
    ratio = 100.0 if in_size == 0 else 100 * out_size / in_size
    msg = (
        f"\nCompression succeeded, {t2 - t1:.2f} seconds.\n"
        f"Input {in_size:,} bytes, output {out_size:,} bytes, ratio {ratio:.2f}%.\n"
    )
    print(msg)


def decompress(args: Args) -> None:
    args.input = cast("BinaryIO", args.input)

    # output file
    if args.output is None:
        if args.test is None:
            from re import subn

            out_path, replaced = subn(r"(?i)^(.*)\.zst$", r"\1", args.input.name)
            if not replaced:
                out_path = args.input.name + ".decompressed"
        else:
            out_path = os.devnull
        open_output(args, out_path)
        args.output = cast("BinaryIO", args.output)

    # option
    option: Mapping[int, int] = {DParameter.windowLogMax: args.windowLogMax}

    # pre-decompress message
    output_name = args.output.name
    if output_name == os.devnull:
        output_name = "None"
    print(
        "Decompress file:\n"
        f" - input file : {args.input.name}\n"
        f" - output file: {output_name}\n"
        f" - zstd dictionary: {args.zd}"
    )

    # decompress
    t1 = time()
    with ZstdFile(args.input, level_or_option=option, zstd_dict=args.zd) as fin:
        copyfileobj(fin, args.output)
    t2 = time()
    in_size = args.input.tell()
    out_size = args.output.tell()
    close_files(args)

    # post-decompress message
    ratio = 100.0 if out_size == 0 else 100 * in_size / out_size
    msg = (
        f"\nDecompression succeeded, {t2 - t1:.2f} seconds.\n"
        f"Input {in_size:,} bytes, output {out_size:,} bytes, ratio {ratio:.2f}%.\n"
    )
    print(msg)


def train(args: Args) -> None:
    from glob import glob

    # check output file
    if args.output is None:
        raise ValueError("need to specify output file using -o/--output option")

    # gather samples
    print("Gathering samples, please wait.", flush=True)
    lst = []
    for file in glob(args.train, recursive=True):
        with open(file, "rb") as f:
            dat = f.read()
            lst.append(dat)
            print("samples count:", len(lst), end="\r", flush=True)
    if len(lst) == 0:
        raise ValueError("No samples gathered, please check GLOB_PATH.")

    samples_size = sum(len(sample) for sample in lst)
    if samples_size == 0:
        raise ValueError("Samples content is empty, can't train.")

    # pre-train message
    msg = (
        "Gathered, train zstd dictionary:\n"
        " - samples: {}\n"
        " - samples number: {}\n"
        " - samples content: {:,} bytes\n"
        " - dict file: {}\n"
        " - dict max size: {:,} bytes\n"
        " - dict id: {}\n"
        "Training, please wait."
    ).format(
        args.train,
        len(lst),
        samples_size,
        args.output.name,
        args.maxdict,
        "random" if args.dictID is None else args.dictID,
    )
    print(msg, flush=True)

    # train
    t1 = time()
    zd = train_dict(lst, args.maxdict)
    t2 = time()

    # Dictionary_ID: 4 bytes, stored in little-endian format.
    # it can be any value, except 0 (which means no Dictionary_ID).
    if args.dictID is not None and len(zd.dict_content) >= 8:
        content = (
            zd.dict_content[:4]
            + args.dictID.to_bytes(4, "little")
            + zd.dict_content[8:]
        )
        zd = ZstdDict(content)

    # save to file
    args.output.write(zd.dict_content)
    close_files(args)

    # post-train message
    msg = f"Training succeeded, {t2 - t1:.2f} seconds.\nDictionary: {zd}\n"
    print(msg)


def tarfile_create(args: Args) -> None:
    import sys

    if sys.version_info < (3, 14):
        from backports.zstd import tarfile
    else:
        import tarfile

    # check input dir
    args.tar_input_dir = args.tar_input_dir.rstrip(os.sep)
    if not os.path.isdir(args.tar_input_dir):
        msg = "Tar archive input dir invalid: " + args.tar_input_dir
        raise NotADirectoryError(msg)
    dirname, basename = os.path.split(args.tar_input_dir)

    # check output file
    if args.output is None:
        out_path = os.path.join(dirname, basename + ".tar.zst")
        open_output(args, out_path)
        args.output = cast("BinaryIO", args.output)

    # pre-compress message
    msg = (
        "Archive tar file:\n"
        f" - input directory: {args.tar_input_dir}\n"
        f" - output file: {args.output.name}"
    )
    print(msg)

    # option
    option = compress_option(args)

    # compress
    print("Archiving, please wait.", flush=True)
    t1 = time()
    with tarfile.TarFile.zstopen(
        None, fileobj=args.output, mode="w", options=option, zstd_dict=args.zd
    ) as f:
        f.add(args.tar_input_dir, basename)
        uncompressed_size = f.fileobj.tell()  # type: ignore[union-attr]
    t2 = time()

    output_file_size = args.output.tell()
    close_files(args)

    # post-compress message
    if uncompressed_size != 0:
        ratio = 100 * output_file_size / uncompressed_size
    else:
        ratio = 100.0

    msg = (
        f"Archiving succeeded, {t2 - t1:.2f} seconds.\n"
        f"Input ~{uncompressed_size:,} bytes, output {output_file_size:,} bytes, ratio {ratio:.2f}%.\n"
    )
    print(msg)


def tarfile_extract(args: Args) -> None:
    import sys

    if sys.version_info < (3, 14):
        from backports.zstd import tarfile
    else:
        import tarfile

    # input file size
    if args.input is None:
        msg = "need to specify input file using -d/--decompress option."
        raise FileNotFoundError(msg)
    input_file_size = os.path.getsize(args.input.name)

    # check output dir
    if not os.path.isdir(args.tar_output_dir):
        msg = "Tar archive output dir invalid: " + args.tar_output_dir
        raise NotADirectoryError(msg)

    # option
    option: Mapping[int, int] = {DParameter.windowLogMax: args.windowLogMax}

    # pre-extract message
    msg = (
        "Extract tar archive:\n"
        f" - input file: {args.input.name}\n"
        f" - output dir: {args.tar_output_dir}\n"
        f" - zstd dictionary: {args.zd}\n"
        "Extracting, please wait."
    )
    print(msg, flush=True)

    # extract
    t1 = time()
    with tarfile.TarFile.zstopen(
        None, fileobj=args.input, mode="r", zstd_dict=args.zd, options=option
    ) as f:
        f.extractall(args.tar_output_dir, filter="data")
        uncompressed_size = f.fileobj.tell()  # type: ignore[union-attr]
    t2 = time()
    close_files(args)

    # post-extract message
    if uncompressed_size != 0:
        ratio = 100 * input_file_size / uncompressed_size
    else:
        ratio = 100.0
    msg = (
        f"Extraction succeeded, {t2 - t1:.2f} seconds.\n"
        f"Input {input_file_size:,} bytes, output ~{uncompressed_size:,} bytes, ratio {ratio:.2f}%.\n"
    )
    print(msg)


def range_action(start: int, end: int) -> type[argparse.Action]:
    class RangeAction(argparse.Action):
        def __call__(
            self,
            _: object,
            namespace: object,
            values: str | Sequence[Any] | None,
            option_string: str | None = None,
        ) -> None:
            # convert to int
            try:
                v = int(values)  # type: ignore[arg-type]
            except ValueError:
                raise TypeError(f"{option_string} should be an integer") from None

            # check range
            if not (start <= v <= end):
                # message
                msg = (
                    f"{option_string} value should: {start} <= v <= {end}. "
                    f"provided value is {v}."
                )
                raise ValueError(msg)

            setattr(namespace, self.dest, v)

    return RangeAction


def parse_arg() -> Args:
    p = argparse.ArgumentParser(
        prog="CLI of pyzstd module",
        description=(
            "The command style is similar to zstd's "
            "CLI, but there are some differences.\n"
            "Zstd's CLI should be faster, it has "
            "some I/O optimizations."
        ),
        epilog=(
            "Examples of use:\n"
            "  compress a file:\n"
            "    python -m pyzstd -c IN_FILE -o OUT_FILE\n"
            "  decompress a file:\n"
            "    python -m pyzstd -d IN_FILE -o OUT_FILE\n"
            "  create a tar archive:\n"
            "    python -m pyzstd --tar-input-dir DIR -o OUT_FILE\n"
            "  extract a tar archive, output will forcibly overwrite existing files:\n"
            "    python -m pyzstd -d IN_FILE --tar-output-dir DIR\n"
            "  train a zstd dictionary, ** traverses sub-directories:\n"
            '    python -m pyzstd --train "E:\\cpython\\**\\*.c" -o OUT_FILE'
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    g = p.add_argument_group("Common arguments")
    g.add_argument(
        "-D",
        "--dict",
        metavar="FILE",
        type=argparse.FileType("rb"),
        help="use FILE as zstd dictionary for compression or decompression",
    )
    g.add_argument(
        "-o", "--output", metavar="FILE", type=str, help="result stored into FILE"
    )
    g.add_argument(
        "-f",
        action="store_true",
        help="disable output check, allows overwriting existing file.",
    )

    g = p.add_argument_group("Compression arguments")
    gm = g.add_mutually_exclusive_group()
    gm.add_argument("-c", "--compress", metavar="FILE", type=str, help="compress FILE")
    gm.add_argument(
        "--tar-input-dir",
        metavar="DIR",
        type=str,
        help=(
            "create a tar archive from DIR. this option overrides -c/--compress option."
        ),
    )
    g.add_argument(
        "-l",
        "--level",
        metavar="#",
        default=compressionLevel_values.default,
        action=range_action(compressionLevel_values.min, compressionLevel_values.max),
        help=f"compression level, range: [{compressionLevel_values.min},{compressionLevel_values.max}], default: {compressionLevel_values.default}.",
    )
    g.add_argument(
        "-t",
        "--threads",
        metavar="#",
        default=0,
        action=range_action(*CParameter.nbWorkers.bounds()),
        help=(
            "spawns # threads to compress. if this option is not "
            "specified or is 0, use single thread mode."
        ),
    )
    g.add_argument(
        "--long",
        metavar="#",
        nargs="?",
        const=27,
        default=-1,
        action=range_action(*CParameter.windowLog.bounds()),
        help="enable long distance matching with given windowLog (default #: 27)",
    )
    g.add_argument(
        "--no-checksum",
        action="store_false",
        dest="checksum",
        default=True,
        help="don't add 4-byte XXH64 checksum to the frame",
    )
    g.add_argument(
        "--no-dictID",
        action="store_false",
        dest="write_dictID",
        default=True,
        help="don't write dictID into frame header (dictionary compression only)",
    )

    g = p.add_argument_group("Decompression arguments")
    gm = g.add_mutually_exclusive_group()
    gm.add_argument(
        "-d", "--decompress", metavar="FILE", type=str, help="decompress FILE"
    )
    g.add_argument(
        "--tar-output-dir",
        metavar="DIR",
        type=str,
        help=(
            "extract tar archive to DIR, "
            "output will forcibly overwrite existing files. "
            "this option overrides -o/--output option."
        ),
    )
    gm.add_argument(
        "--test",
        metavar="FILE",
        type=str,
        help="try to decompress FILE to check integrity",
    )
    g.add_argument(
        "--windowLogMax",
        metavar="#",
        default=0,
        action=range_action(*DParameter.windowLogMax.bounds()),
        help="set a memory usage limit for decompression (windowLogMax)",
    )

    g = p.add_argument_group("Dictionary builder")
    g.add_argument(
        "--train",
        metavar="GLOB_PATH",
        type=str,
        help="create a dictionary from a training set of files",
    )
    g.add_argument(
        "--maxdict",
        metavar="SIZE",
        type=int,
        default=112640,
        help="limit dictionary to SIZE bytes (default: 112640)",
    )
    g.add_argument(
        "--dictID",
        metavar="DICT_ID",
        default=None,
        action=range_action(1, 0xFFFFFFFF),
        help="specify dictionary ID value (default: random)",
    )

    args = p.parse_args()

    # input file
    if args.compress is not None:
        args.input = open(args.compress, "rb", buffering=C_READ_BUFFER)  # noqa: SIM115
    elif args.decompress is not None:
        args.input = open(args.decompress, "rb", buffering=D_READ_BUFFER)  # noqa: SIM115
    elif args.test is not None:
        args.input = open(args.test, "rb", buffering=D_READ_BUFFER)  # noqa: SIM115
    else:
        args.input = None

    # output file
    if args.output is not None:
        open_output(args, args.output)

    # load dictionary
    if args.dict is not None:
        zd_content = args.dict.read()
        args.dict.close()
        # Magic_Number: 4 bytes, value 0xEC30A437, little-endian format.
        is_raw = zd_content[:4] != b"\x37\xa4\x30\xec"
        args.zd = ZstdDict(zd_content, is_raw=is_raw)
    else:
        args.zd = None

    # arguments combination
    functions = [
        args.compress,
        args.decompress,
        args.test,
        args.train,
        args.tar_input_dir,
    ]
    if sum(1 for i in functions if i is not None) > 1:
        raise ValueError("Wrong arguments combination")

    return args


def main() -> None:
    print(f"*** pyzstd module v{pyzstd_version}, zstd library v{zstd_version}. ***\n")

    args = parse_arg()

    if args.tar_input_dir:
        tarfile_create(args)
    elif args.tar_output_dir:
        tarfile_extract(args)
    elif args.compress:
        compress(args)
    elif args.decompress or args.test:
        decompress(args)
    elif args.train:
        train(args)
    else:
        print("Invalid command. See help: python -m pyzstd --help")


if __name__ == "__main__":
    main()
